"""Test Text Transformers."""

import re
from string import ascii_uppercase
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest

from rdt.errors import TransformerProcessingError
from rdt.transformers.text import RegexGenerator


class AsciiGenerator:
    """Ascii Upercase Generator."""

    def __init__(self, max_size=26):
        self.pos = 0
        self.max_size = max_size

    def __iter__(self):
        return self

    def __next__(self):
        if self.pos >= self.max_size:
            raise StopIteration

        char = ascii_uppercase[self.pos]
        self.pos += 1

        return char


class TestRegexGenerator:
    """Test class for ``RegexGenerator``."""

    def test___getstate__(self):
        """Test that ``__getstate__`` returns a dictionary without the generator."""
        # Setup
        instance = RegexGenerator()
        instance.reset_randomization()
        mock_random_sates = Mock()
        instance.random_states = mock_random_sates

        # Run
        state = instance.__getstate__()

        # Assert
        assert state == {
            'data_length': None,
            'enforce_uniqueness': False,
            'generated': 0,
            'generator_size': 380204032,
            'output_properties': {None: {'next_transformer': None}},
            'regex_format': '[A-Za-z]{5}',
            'random_states': mock_random_sates,
        }

    @patch('rdt.transformers.text.strings_from_regex')
    def test___setstate__generated_and_generator_size(self, mock_strings_from_regex):
        """Test that ``__setstate__`` will initialize a generator and wind it forward."""
        # Setup
        state = {
            'data_length': None,
            'enforce_uniqueness': False,
            'generated': 10,
            'generator_size': 380204032,
            'output_properties': {None: {'next_transformer': None}},
            'regex_format': '[A-Za-z]{5}'
        }
        generator = AsciiGenerator()
        mock_strings_from_regex.return_value = (generator, 26)
        instance = RegexGenerator()

        # Run
        instance.__setstate__(state)

        # Assert
        assert next(generator) == 'K'
        assert instance.generated == 10
        assert instance.generator_size == 380204032
        mock_strings_from_regex.assert_called_once_with('[A-Za-z]{5}')

    @patch('rdt.transformers.text.strings_from_regex')
    def test___setstate__(self, mock_strings_from_regex):
        """Test that ``__setstate__`` will initialize a generator but not forward it.

        When ``generated`` is ``None`` and ``generator_size`` is ``None`` this will be assigned
        the ``0`` and the ``generator_size`` respectively.
        """
        # Setup
        state = {
            'data_length': None,
            'enforce_uniqueness': False,
            'generated': None,
            'generator_size': None,
            'output_properties': {None: {'next_transformer': None}},
            'regex_format': '[A-Za-z]{5}'
        }
        generator = AsciiGenerator()
        mock_strings_from_regex.return_value = (generator, 26)
        instance = RegexGenerator()

        # Run
        instance.__setstate__(state)

        # Assert
        assert next(generator) == 'A'
        assert instance.generated == 0
        assert instance.generator_size == 26
        mock_strings_from_regex.assert_called_once_with('[A-Za-z]{5}')

    def test___init__default(self):
        """Test the default instantiation of the transformer.

        Test that ``RegexGenerator`` defaults to ``regex_format='[A-Za-z]{5}'``

        Side effects:
            - the ``instance.regex_format`` is ``'[A-Za-z]{5}'``'.
        """
        # Run
        instance = RegexGenerator()

        # Assert
        assert instance.data_length is None
        assert instance.regex_format == '[A-Za-z]{5}'
        assert instance.enforce_uniqueness is False

    def test___init__custom(self):
        """Test the default instantiation of the transformer.

        Test that when creating an instance of ``RegexGenerator`` and passing a
        ``regex_format`` this is being stored.

        Side effects:
            - the ``instance.regex_format`` is ``'[A-Za-z]{5}'``'.
            - ``instance.enforce_uniqueness`` is ``True``.
        """
        # Run
        instance = RegexGenerator(
            regex_format='[0-9]',
            enforce_uniqueness=True
        )

        # Assert
        assert instance.data_length is None
        assert instance.regex_format == '[0-9]'
        assert instance.enforce_uniqueness

    @patch('rdt.transformers.text.BaseTransformer.reset_randomization')
    @patch('rdt.transformers.text.strings_from_regex')
    def test_reset_randomization(self, mock_strings_from_regex, mock_base_reset):
        """Test that this method creates a new generator.

        This method should create a new ``instance.generator``, ``instance.generator_size`` and
        restart the ``instance.generated`` values to 0.
        """
        # Setup
        generator = AsciiGenerator(5)
        mock_strings_from_regex.return_value = (generator, 2)
        instance = RegexGenerator()

        # Run
        instance.reset_randomization()

        # Assert
        assert instance.generator == generator
        assert instance.generator_size == 2
        assert instance.generated == 0
        mock_strings_from_regex.assert_called_once_with('[A-Za-z]{5}')
        mock_base_reset.assert_called_once()

    def test__fit(self):
        """Test the ``_fit`` method.

        Validate that the ``_fit`` method learns the original data length.

        Setup:
            - Initialize a ``RegexGenerator`` transformer.

        Input:
            - ``pd.Series`` containing 3 strings.

        Side Effects:
            - ``instance.data_length`` equals to the length of the input data.
        """
        # Setup
        instance = RegexGenerator()
        columns_data = pd.Series(['1', '2', '3'])

        # Run
        instance._fit(columns_data)

        # Assert
        assert instance.data_length == 3
        assert instance.output_properties == {None: {'next_transformer': None}}

    def test__transform(self):
        """Test the ``_transform`` method.

        Validate that the ``_transform`` method returns ``None``.

        Setup:
            - Initialize a ``RegexGenerator`` transformer.

        Input:
            - ``pd.Series`` with three values.

        Output:
            - ``None``.
        """
        # Setup
        columns_data = pd.Series([1, 2, 3])
        instance = RegexGenerator()

        # Run
        result = instance._transform(columns_data)

        # Assert
        assert result is None

    def test__reverse_transform_generator_size_bigger_than_data_length(self):
        """Test the ``_reverse_transform`` method.

        Validate that the ``_reverse_transform`` method uses the ``instance.generator``
        to generate the ``instance.data_length`` number of data.

        Setup:
            - Initialize a ``RegexGenerator`` instance.
            - Set ``data_length`` to 3.
            - Initialize a generator.
            - Set a generator, generator size and generated values.

        Output:
            - A ``numpy.array`` with the first three letters from the generator.
        """
        # Setup
        instance = RegexGenerator('[A-Z]')
        columns_data = pd.Series()
        instance.data_length = 3
        generator = AsciiGenerator(max_size=5)
        instance.generator = generator
        instance.generator_size = 5
        instance.generated = 0

        # Run
        result = instance._reverse_transform(columns_data)

        # Assert
        np.testing.assert_array_equal(result, np.array(['A', 'B', 'C']))

    def test__reverse_transform_generator_size_smaller_than_data_length(self):
        """Test the ``_reverse_transform`` method.

        Validate that the ``_reverse_transform`` method uses the ``instance.generator``
        to generate the ``instance.data_length`` number of data when ``enforce_uniqueness`` is
        ``False`` but the data to be created is bigger.

        Setup:
            - Initialize a ``RegexGenerator`` instance.
            - Set ``data_length`` to 11.
            - Initialize a generator.

        Output:
            - A ``numpy.array`` with the first five letters from the generator repeated.
        """
        # Setup
        instance = RegexGenerator('[A-Z]', enforce_uniqueness=False)
        columns_data = pd.Series()
        instance.reset_randomization = Mock()
        instance.data_length = 11
        generator = AsciiGenerator(5)
        instance.columns = ['a']
        instance.generator = generator
        instance.generator_size = 5
        instance.generated = 0

        # Run
        result = instance._reverse_transform(columns_data)

        # Assert
        expected_result = np.array(['A', 'B', 'C', 'D', 'E', 'A', 'B', 'C', 'D', 'E', 'A'])
        np.testing.assert_array_equal(result, expected_result)

    def test__reverse_transform_generator_size_of_input_data(self):
        """Test the ``_reverse_transform`` method.

        Validate that the ``_reverse_transform`` method uses the ``instance.generator``
        to generate the ``instance.data_length`` number of data when ``enforce_uniqueness`` is
        ``False`` but the data to be created is bigger.

        Setup:
            - Initialize a ``RegexGenerator`` instance.
            - Set ``data_length`` to 2.
            - Initialize a generator.

        Input:
            - ``pandas.Series`` with a length of ``4``.

        Output:
            - A ``numpy.array`` with the first five letters from the generator repeated.
        """
        # Setup
        instance = RegexGenerator('[A-Z]')
        columns_data = pd.Series([1, 2, 3, 4])
        instance.data_length = 2
        generator = AsciiGenerator(5)
        instance.generator = generator
        instance.generator_size = 5
        instance.generated = 0
        instance.columns = ['a']

        # Run
        result = instance._reverse_transform(columns_data)

        # Assert
        expected_result = np.array(['A', 'B', 'C', 'D'])
        np.testing.assert_array_equal(result, expected_result)
        assert instance.generated == 4

    def test__reverse_transform_not_enough_unique_values(self):
        """Test the ``_reverse_transform`` method.

        Validate that the ``_reverse_transform`` method calls the ``strings_from_regex``
        function using the ``instance.regex_format`` and then generates the
        ``instance.data_length`` number of data.

        Setup:
            - Initialize a ``RegexGenerator`` instance with ``enforce_uniqueness`` to ``True``.
            - Set ``data_length`` to 6.
            - Initialize a generator.

        Side Effects:
            - An error is being raised as not enough unique values can be generated.
        """
        # Setup
        instance = RegexGenerator('[A-Z]', enforce_uniqueness=True)
        instance.data_length = 6
        generator = AsciiGenerator(5)
        instance.generator = generator
        instance.generator_size = 5
        instance.generated = 0
        instance.columns = ['a']
        columns_data = pd.Series()

        # Assert
        error_msg = re.escape(
            'The regex is not able to generate 6 unique values. Please use a different regex '
            "for column ('a')."
        )
        with pytest.raises(TransformerProcessingError, match=error_msg):
            instance._reverse_transform(columns_data)

    def test__reverse_transform_enforce_uniqueness_not_enough_remaining(self):
        """Test the case when ``_reverse_transform`` can't generate enough new values.

        Validate that an error is being raised stating that not enough new values can be
        generated and that the user can use ``reset_randomization`` in order to restart the
        generator.
        """
        # Setup
        instance = RegexGenerator('[A-Z]', enforce_uniqueness=True)
        instance.data_length = 6
        generator = AsciiGenerator(10)
        instance.generator = generator
        instance.generator_size = 10
        instance.generated = 9
        instance.columns = ['a']
        columns_data = pd.Series()

        # Assert
        error_msg = re.escape(
            'The regex generator is not able to generate 6 new unique values (only 1 unique '
            "value left). Please use 'reset_randomization' in order to restart the generator."
        )
        with pytest.raises(TransformerProcessingError, match=error_msg):
            instance._reverse_transform(columns_data)



"""Integration tests for the HyperTransformer."""

import re

import numpy as np
import pandas as pd
import pytest

from rdt import HyperTransformer, get_demo
from rdt.errors import ConfigNotSetError, InvalidConfigError, InvalidDataError, NotFittedError
from rdt.transformers import (
    AnonymizedFaker, BaseTransformer, BinaryEncoder, ClusterBasedNormalizer, FloatFormatter,
    FrequencyEncoder, LabelEncoder, OneHotEncoder, RegexGenerator, UnixTimestampEncoder,
    get_default_transformer, get_default_transformers)
from rdt.transformers.datetime import OptimizedTimestampEncoder
from rdt.transformers.numerical import GaussianNormalizer
from rdt.transformers.pii.anonymizer import PseudoAnonymizedFaker


class DummyTransformerNumerical(BaseTransformer):

    INPUT_SDTYPE = 'categorical'

    def _fit(self, data):
        pass

    def _transform(self, data):
        return data.astype(float)

    def _reverse_transform(self, data):
        return data.astype(str)


class DummyTransformerNotMLReady(BaseTransformer):

    INPUT_SDTYPE = 'datetime'

    def __init__(self):
        super().__init__()
        self.output_properties = {
            None: {'sdtype': 'datetime', 'next_transformer': FrequencyEncoder()}
        }

    def _fit(self, data):
        pass

    def _transform(self, data):
        # Stringify input data
        return data.astype(str)

    def _reverse_transform(self, data):
        return data.astype('datetime64[ns]')


TEST_DATA_INDEX = [4, 6, 3, 8, 'a', 1.0, 2.0, 3.0]


def get_input_data():
    datetimes = pd.to_datetime([
        '2010-02-01',
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
        '2010-01-01',
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
    ])
    data = pd.DataFrame({
        'integer': [1, 2, 1, 3, 1, 4, 2, 3],
        'float': [0.1, 0.2, 0.1, 0.2, 0.1, 0.4, 0.2, 0.3],
        'categorical': ['a', 'a', 'b', 'b', 'a', 'b', 'a', 'a'],
        'bool': [False, False, False, True, False, False, True, False],
        'datetime': datetimes,
        'names': ['Jon', 'Arya', 'Arya', 'Jon', 'Jon', 'Sansa', 'Jon', 'Jon'],
    }, index=TEST_DATA_INDEX)

    return data


def get_transformed_data():
    datetimes = [
        1.264982e+18,
        1.264982e+18,
        1.262304e+18,
        1.262304e+18,
        1.262304e+18,
        1.264982e+18,
        1.262304e+18,
        1.262304e+18
    ]
    return pd.DataFrame({
        'integer': [1., 2., 1., 3., 1., 4., 2., 3.],
        'float': [0.1, 0.2, 0.1, 0.2, 0.1, 0.4, 0.2, 0.3],
        'categorical': [
            0.9690758764963199, 0.8816575994729887, 1.1326495454234662, 1.7988488918189502,
            0.9265972159030215, 1.885454600378942, 0.9280858691537548, 0.5093227924068265
        ],
        'bool': [
            0.26161253184788935, 0.5735484647493089, 0.026673806296574787, 1.197229599974477,
            0.8860641570557322, 0.33432787358513416, 1.1089412122841389, 0.6182653878449814
        ],
        'datetime': datetimes,
        'names': [
            0.24180193241041126, 1.9297787196579723, 1.5617500744772101, 0.6811042561384157,
            0.48017218468846856, 2.2867787591284823, 0.25476586891248476, 0.620052082101593
        ]
    }, index=TEST_DATA_INDEX)


def get_reversed_data():
    data = get_input_data()

    return data


def test_hypertransformer_default_inputs():
    """Test the HyperTransformer with default parameters.

    This tests that if default parameters are provided to the HyperTransformer,
    the ``default_transformers`` method will be used to determine which
    transformers to use for each field.

    Input:
        - A dataframe with every sdtype.
        - A fixed random seed to guarantee the samle values are null.

    Expected behavior:
        - The transformed data should contain all the ML ready data.
        - The reverse transformed data should be the same as the input.
    """
    # Setup
    datetimes = pd.to_datetime([
        np.nan,
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
        '2010-01-01',
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
    ])
    data = pd.DataFrame({
        'integer': [1, 2, 1, 3, 1, 4, 2, 3],
        'float': [0.1, 0.2, 0.1, np.nan, 0.1, 0.4, np.nan, 0.3],
        'categorical': ['a', 'a', np.nan, 'b', 'a', 'b', 'a', 'a'],
        'bool': [False, np.nan, False, True, False, np.nan, True, False],
        'datetime': datetimes,
        'names': ['Jon', 'Arya', 'Arya', 'Jon', 'Jon', 'Sansa', 'Jon', 'Jon'],
    }, index=TEST_DATA_INDEX)

    # Run
    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.fit(data)
    transformed = ht.transform(data)
    reverse_transformed = ht.reverse_transform(transformed)

    # Assert
    expected_datetimes = [
        1.263069e+18,
        1.264982e+18,
        1.262304e+18,
        1.262304e+18,
        1.262304e+18,
        1.264982e+18,
        1.262304e+18,
        1.262304e+18
    ]
    expected_transformed = pd.DataFrame({
        'integer': [1., 2., 1., 3., 1., 4., 2., 3.],
        'float': [0.1, 0.2, 0.1, 0.2, 0.1, 0.4, 0.2, 0.3],
        'categorical': [
            0.9690758764963199, 0.8816575994729887, 1.1326495454234662, 2.7988488918189502,
            0.9265972159030215, 2.8854546003789423, 0.9280858691537548, 0.5093227924068265
        ],
        'bool': [
            0.26161253184788935, 1.573548464749309, 0.026673806296574787, 2.1972295999744773,
            0.8860641570557322, 1.334327873585134, 2.108941212284139, 0.6182653878449814
        ],
        'datetime': expected_datetimes,
        'names': [
            0.24180193241041126, 1.9297787196579723, 1.5617500744772101, 0.6811042561384157,
            0.48017218468846856, 2.2867787591284823, 0.25476586891248476, 0.620052082101593
        ]
    }, index=TEST_DATA_INDEX)
    pd.testing.assert_frame_equal(transformed, expected_transformed)

    reversed_datetimes = pd.to_datetime([
        '2010-01-09',
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
        '2010-01-01',
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
    ])
    expected_reversed = pd.DataFrame({
        'integer': [1, 2, 1, 3, 1, 4, 2, 3],
        'float': [0.1, 0.2, 0.1, 0.20000000000000004, 0.1, 0.4, 0.20000000000000004, 0.3],
        'categorical': ['a', 'a', np.nan, 'b', 'a', 'b', 'a', 'a'],
        'bool': [False, False, False, True, False, False, True, False],
        'datetime': reversed_datetimes,
        'names': ['Jon', 'Arya', 'Arya', 'Jon', 'Jon', 'Sansa', 'Jon', 'Jon'],
    }, index=TEST_DATA_INDEX)
    for row in range(reverse_transformed.shape[0]):
        for column in range(reverse_transformed.shape[1]):
            expected = expected_reversed.iloc[row, column]
            actual = reverse_transformed.iloc[row, column]
            assert pd.isna(actual) or expected == actual

    assert isinstance(ht.field_transformers['integer'], FloatFormatter)
    assert isinstance(ht.field_transformers['float'], FloatFormatter)
    assert isinstance(ht.field_transformers['categorical'], LabelEncoder)
    assert isinstance(ht.field_transformers['bool'], LabelEncoder)
    assert isinstance(ht.field_transformers['datetime'], UnixTimestampEncoder)
    assert isinstance(ht.field_transformers['names'], LabelEncoder)

    get_default_transformers.cache_clear()
    get_default_transformer.cache_clear()


def test_hypertransformer_field_transformers():
    """Test the HyperTransformer with ``field_transformers`` provided.

    This tests that the transformers specified in the ``field_transformers``
    argument are used. Any output of a transformer that is not ML ready (not
    in the ``_transform_output_sdtypes`` list) should be recursively transformed
    till it is.

    Setup:
        - The datetime column is set to use a dummy transformer that stringifies
        the input. That output is then set to use the categorical transformer.

    Input:
        - A dict mapping each field to a transformer.
        - A dataframe with every sdtype.

    Expected behavior:
        - The transformed data should contain all the ML ready data.
        - The reverse transformed data should be the same as the input.
    """
    # Setup
    config = {
        'sdtypes': {
            'integer': 'numerical',
            'float': 'numerical',
            'categorical': 'categorical',
            'bool': 'boolean',
            'datetime': 'datetime',
            'names': 'categorical'
        },
        'transformers': {
            'integer': FloatFormatter(missing_value_replacement='mean'),
            'float': FloatFormatter(missing_value_replacement='mean'),
            'categorical': LabelEncoder(add_noise=True),
            'bool': LabelEncoder(add_noise=True),
            'datetime': DummyTransformerNotMLReady(),
            'names': LabelEncoder(add_noise=True)
        }
    }

    data = get_input_data()

    # Run
    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.set_config(config)
    ht.fit(data)
    transformed = ht.transform(data)
    reverse_transformed = ht.reverse_transform(transformed)

    # Assert
    expected_transformed = get_transformed_data()
    transformed_datetimes = [0.8125, 0.8125, 0.3125, 0.3125, 0.3125, 0.8125, 0.3125, 0.3125]
    expected_transformed['datetime'] = transformed_datetimes
    pd.testing.assert_frame_equal(transformed, expected_transformed)

    expected_reversed = get_reversed_data()
    pd.testing.assert_frame_equal(expected_reversed, reverse_transformed)


def test_single_category():
    """Test that categorical variables with a single value are supported."""
    # Setup
    ht = HyperTransformer()
    data = pd.DataFrame({
        'a': ['a', 'a', 'a']
    })

    # Run
    ht.detect_initial_config(data)
    ht.update_transformers(column_name_to_transformer={
        'a': OneHotEncoder()
    })
    ht.fit(data)
    transformed = ht.transform(data)
    reverse = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(data, reverse)


def test_categorical_encoders_with_booleans():
    """Test that categorical encoders support boolean values."""
    # Setup
    config = {
        'sdtypes': {
            'email_confirmed': 'boolean',
            'subscribed': 'boolean',
            'paid': 'boolean',
        },
        'transformers': {
            'email_confirmed': FrequencyEncoder(),
            'subscribed': OneHotEncoder(),
            'paid': LabelEncoder(),
        }
    }

    ht = HyperTransformer()

    # Run and Assert
    ht.set_config(config)


def test_dtype_category():
    """Test that categorical variables of dtype category are supported."""
    # Setup
    data = pd.DataFrame({'a': ['a', 'b', 'c']}, dtype='category')

    # Run
    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.fit(data)
    transformed = ht.transform(data)
    reverse = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(reverse, data)


def test_multiple_fits():
    """HyperTransformer should be able to be used multiple times.

    Fitting, transforming and reverse transforming should produce the same results when
    called on the same data multiple times.
    """
    # Setup
    data = get_input_data()
    ht = HyperTransformer()

    # Run
    ht.detect_initial_config(data)
    ht.fit(data)
    transformed1 = ht.transform(data)
    reversed1 = ht.reverse_transform(transformed1)

    ht.detect_initial_config(data)
    ht.fit(data)
    transformed2 = ht.transform(data)
    reversed2 = ht.reverse_transform(transformed2)

    # Assert
    pd.testing.assert_frame_equal(transformed1, transformed2)
    pd.testing.assert_frame_equal(reversed1, reversed2)


def test_multiple_fits_different_data():
    """HyperTransformer should be able to be used multiple times regardless of the data.

    Fitting, transforming and reverse transforming should work when called on different data.
    """
    # Setup
    data = pd.DataFrame({'col1': [1, 2, 3], 'col2': [1.0, 0.0, 0.0]})
    new_data = pd.DataFrame({'col2': [1, 2, 3], 'col1': [1.0, 0.0, 0.0]})
    ht = HyperTransformer()

    # Run
    ht.detect_initial_config(data)
    ht.fit(data)
    ht.detect_initial_config(new_data)
    ht.fit(new_data)
    transformed1 = ht.transform(new_data)
    transformed2 = ht.transform(new_data)
    reverse1 = ht.reverse_transform(transformed1)
    reverse2 = ht.reverse_transform(transformed2)

    # Assert
    expected_transformed = pd.DataFrame(
        {'col2': [1., 2., 3.], 'col1': [1.0, 0.0, 0.0]})
    pd.testing.assert_frame_equal(transformed1, expected_transformed)
    pd.testing.assert_frame_equal(transformed2, expected_transformed)
    pd.testing.assert_frame_equal(reverse1, new_data)
    pd.testing.assert_frame_equal(reverse2, new_data)


def test_multiple_fits_different_columns():
    """HyperTransformer should be able to be used multiple times regardless of the data.

    Fitting, transforming and reverse transforming should work when called on different data.
    """
    # Setup
    data = pd.DataFrame({'col1': [1, 2, 3], 'col2': [1.0, 0.0, 0.0]})
    new_data = pd.DataFrame({'col3': [1, 2, 3], 'col4': [1.0, 0.0, 0.0]})
    ht = HyperTransformer()

    # Run
    ht.detect_initial_config(data)
    ht.fit(data)
    ht.detect_initial_config(new_data)
    ht.fit(new_data)
    transformed1 = ht.transform(new_data)
    transformed2 = ht.transform(new_data)
    reverse1 = ht.reverse_transform(transformed1)
    reverse2 = ht.reverse_transform(transformed2)

    # Assert
    expected_transformed = pd.DataFrame(
        {'col3': [1., 2., 3.], 'col4': [1.0, 0.0, 0.0]})
    pd.testing.assert_frame_equal(transformed1, expected_transformed)
    pd.testing.assert_frame_equal(transformed2, expected_transformed)
    pd.testing.assert_frame_equal(reverse1, new_data)
    pd.testing.assert_frame_equal(reverse2, new_data)


def test_multiple_fits_with_set_config():
    """HyperTransformer should be able to be used multiple times regardless of the data.

    Fitting, transforming and reverse transforming should work when called on different data.
    """
    # Setup
    data = get_input_data()
    ht = HyperTransformer()

    # Run
    ht.detect_initial_config(data)
    ht.set_config(config={
        'sdtypes': {'integer': 'categorical'},
        'transformers': {'integer': FrequencyEncoder()}
    })
    ht.fit(data)
    transformed1 = ht.transform(data)
    reverse1 = ht.reverse_transform(transformed1)

    ht.fit(data)
    transformed2 = ht.transform(data)
    reverse2 = ht.reverse_transform(transformed2)

    # Assert
    pd.testing.assert_frame_equal(transformed1, transformed2)
    pd.testing.assert_frame_equal(reverse1, reverse2)


def test_multiple_detect_configs_with_set_config():
    """HyperTransformer should be able to be used multiple times regardless of the data.

    Fitting, transforming and reverse transforming should work when called on different data.
    """
    # Setup
    data = get_input_data()
    ht = HyperTransformer()

    # Run
    ht.detect_initial_config(data)
    ht.fit(data)
    transformed1 = ht.transform(data)
    reverse1 = ht.reverse_transform(transformed1)

    ht.set_config(config={
        'sdtypes': {'integers': 'categorical'},
        'transformers': {'integers': FrequencyEncoder()}
    })

    ht.detect_initial_config(data)
    ht.fit(data)
    transformed2 = ht.transform(data)
    reverse2 = ht.reverse_transform(transformed2)

    # Assert
    pd.testing.assert_frame_equal(transformed1, transformed2)
    pd.testing.assert_frame_equal(reverse1, reverse2)


def test_detect_initial_config_doesnt_affect_fit():
    """HyperTransformer should fit the same way regardless of ``detect_initial_config``.

    Calling the ``detect_initial_config`` method should not affect the results of ``fit``,
    ``transform`` or ``reverse_transform``.
    """
    # Setup
    data = get_input_data()
    ht = HyperTransformer()

    # Run
    ht.detect_initial_config(data)
    ht.fit(data)
    transformed1 = ht.transform(data)
    reversed1 = ht.reverse_transform(transformed1)

    ht.detect_initial_config(data)
    ht.fit(data)
    transformed2 = ht.transform(data)
    reversed2 = ht.reverse_transform(transformed1)

    # Assert
    pd.testing.assert_frame_equal(transformed1, transformed2)
    pd.testing.assert_frame_equal(reversed1, reversed2)


def test_multiple_detects():
    """HyperTransformer should be able to be used multiple times regardless of the data.

    Fitting, transforming and reverse transforming should work when called on different data.
    """
    # Setup
    data = pd.DataFrame({'col2': [1, 2, 3], 'col1': [1.0, 0.0, 0.0]})
    new_data = get_input_data()
    ht = HyperTransformer()

    # Run
    ht.detect_initial_config(data)
    ht.detect_initial_config(new_data)
    ht.fit(new_data)
    transformed = ht.transform(new_data)
    reverse = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(transformed, get_transformed_data())
    pd.testing.assert_frame_equal(reverse, get_reversed_data())


def test_transform_without_fit():
    """HyperTransformer should raise an error when transforming without fitting."""
    # Setup
    data = pd.DataFrame({'column': [1, 2, 3]})
    ht = HyperTransformer()
    ht.detect_initial_config(data)

    # Run / Assert
    with pytest.raises(NotFittedError):
        ht.transform(data)


def test_fit_data_different_than_detect():
    """HyperTransformer should raise an error when transforming without fitting."""
    # Setup
    ht = HyperTransformer()
    detect_data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    data = pd.DataFrame({'col1': [1, 2], 'col3': ['a', 'b']})

    # Run / Assert
    error_msg = re.escape(
        'The data you are trying to fit has different columns than the original '
        "detected data (unknown columns: ['col3']). Column names and their "
        "sdtypes must be the same. Use the method 'get_config()' to see the expected "
        'values.'
    )
    ht.detect_initial_config(detect_data)
    with pytest.raises(InvalidDataError, match=error_msg):
        ht.fit(data)


def test_transform_without_fitting():
    """HyperTransformer shouldn't transform when fit hasn't been called yet."""
    # Setup
    data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    ht = HyperTransformer()

    # Run / Assert
    ht.detect_initial_config(data)
    error_msg = (
        'The HyperTransformer is not ready to use. Please fit your data first using '
        "'fit' or 'fit_transform'."
    )
    with pytest.raises(NotFittedError, match=error_msg):
        ht.transform(data)


def test_transform_without_refitting():
    """HyperTransformer shouldn't transform when a new config hasn't been fitted."""
    # Setup
    data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    ht = HyperTransformer()

    # Run / Assert
    ht.detect_initial_config(data)
    ht.fit(data)
    ht.update_sdtypes({'col1': 'categorical'})
    error_msg = (
        'The HyperTransformer is not ready to use. Please fit your data first using '
        "'fit' or 'fit_transform'."
    )
    with pytest.raises(NotFittedError, match=error_msg):
        ht.transform(data)


def test_transform_without_config():
    """HyperTransformer shouldn't transform when a config hasn't been set."""
    # Setup
    data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    ht = HyperTransformer()

    # Run / Assert
    error_msg = (
        "No config detected. Set the config using 'set_config' or pre-populate "
        "it automatically from your data using 'detect_initial_config' prior to "
        'fitting your data.'
    )
    with pytest.raises(ConfigNotSetError, match=error_msg):
        ht.transform(data)


def test_transform_unseen_columns():
    """HyperTransformer shouldn't transform when the data wasn't seen during fit."""
    # Setup
    data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    different_data = pd.DataFrame({'col3': [1, 2]})
    ht = HyperTransformer()

    # Run / Assert
    ht.detect_initial_config(data)
    ht.fit(data)
    error_msg = error_msg = (
        'The data you are trying to transform has different columns than the original data. '
        'Column names and their sdtypes must be the same.'
    )
    with pytest.raises(InvalidDataError, match=error_msg):
        ht.transform(different_data)


def test_update_sdtypes_incorrect_columns():
    """HyperTransformer should crash when update_sdytpes is passed non-existing columns."""
    # Setup
    data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    column_name_to_sdtype = {'col3': [1, 2]}
    ht = HyperTransformer()

    # Run / Assert
    ht.detect_initial_config(data)
    error_msg = error_msg = re.escape(
        "Invalid column names: ['col3']. These columns do not exist in the "
        "config. Use 'set_config()' to write and set your entire config at once."
    )
    with pytest.raises(InvalidConfigError, match=error_msg):
        ht.update_sdtypes(column_name_to_sdtype)


def test_update_sdtypes_incorrect_sdtype():
    """HyperTransformer should crash when update_sdytpes is passed non-existing columns."""
    # Setup
    data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    column_name_to_sdtype = {'col1': 'unexpected'}
    ht = HyperTransformer()

    # Run / Assert
    ht.detect_initial_config(data)
    error_msg = error_msg = re.escape(
        "Invalid sdtypes: ['unexpected']. If you are trying to use a "
        'premium sdtype, contact info@sdv.dev about RDT Add-Ons.'
    )
    with pytest.raises(InvalidConfigError, match=error_msg):
        ht.update_sdtypes(column_name_to_sdtype)


def test_transform_subset():
    """Test the ``transform_subset`` method.

    The method should return a ``pandas.DataFrame`` with the subset of columns transformed.

    Setup:
        - Detect the config and fit the data.

    Input:
        - A ``pandas.DataFrame`` with a subset of the fitted columns.

    Ouput:
        - A ``pandas.DataFrame`` with the subset transformed.
    """
    # Setup
    data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    subset = data[['col1']]
    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.fit(data)

    # Run
    transformed = ht.transform_subset(subset)

    # Assert
    expected = pd.DataFrame({'col1': [1., 2.]})
    pd.testing.assert_frame_equal(transformed, expected)


def test_reverse_transform_subset():
    """Test the ``reverse_transform_subset`` method.

    The method should return a ``pandas.DataFrame`` with the subset of columns reverse transformed.

    Setup:
        - Detect the config and fit the data.

    Input:
        - A ``pandas.DataFrame`` with a subset of the output columns.

    Ouput:
        - A ``pandas.DataFrame`` with the subset reverse transformed.
    """
    # Setup
    data = pd.DataFrame({'col1': [1, 2], 'col2': ['a', 'b']})
    subset = pd.DataFrame({'col1': [1, 2]})
    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.fit(data)

    # Run
    reverse_transformed = ht.reverse_transform_subset(subset)

    # Assert
    expected = pd.DataFrame({'col1': [1, 2]})
    pd.testing.assert_frame_equal(reverse_transformed, expected)


def test_hyper_transformer_with_supported_sdtypes():
    """Test the ``HyperTransformer`` that supports multiple ``sdtypes`` for a ``Transformer``.

    Test that the ``HyperTransformer`` works with ``get_supported_sdtypes`` allowing us
    to asign different transformer to a ``sdtype``. For example, a ``FrequencyEncoder`` to
    a ``boolean`` sdtype.

    Setup:
        - Dataframe with multiple datatypes.
        - Instance of ``HyperTransformer``.
        - Update the transformer for ``boolean`` sdtype to ``FrequencyEncoder()``.

    Run:
        - Run end to end the ``hypertransformer``.

    Assert:
        - Assert that the ``FerquencyEncoder`` is used for the ``boolean`` data.
    """
    # Setup
    data = pd.DataFrame({
        'user': ['John', 'Doe', 'John Doe', 'Doe John'],
        'id': list(range(4)),
        'subscribed': [True, False, True, False]
    })

    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.update_transformers_by_sdtype(
        sdtype='boolean',
        transformer=FrequencyEncoder(add_noise=True)
    )

    # Run
    transformed = ht.fit_transform(data)
    ht.reverse_transform(transformed)

    for transformer in ht.get_config()['transformers'].values():
        assert not isinstance(transformer, BinaryEncoder)


def test_hyper_transformer_reverse_transform_subset_and_generators():
    """Test the ``HyperTransformer`` with ``reverse_transform_subset``.

    Test that when calling ``reverse_transform_subset`` and there are ``generators`` like
    ``AnonymizedFaker`` or ``RegexGenerator`` those are not being used in the ``subset``, and
    also any other transformer which can't transform the given columns.

    Setup:
        - DataFrame with multiple datatypes.
        - Instance of HyperTransformer.
        - Add ``pii`` using ``AnonymizedFaker`` and ``RegexGenerator``.

    Run:
        - Use ``fit_transform`` then ``revese_transform_subsample``.

    Assert:
        - Assert that the ``reverse_transformed`` data does not contain any additional columns
          but the expected one.
    """
    # Setup
    customers = get_demo()
    customers['id'] = ['ID_a', 'ID_b', 'ID_c', 'ID_d', 'ID_e']

    # Create a config
    ht = HyperTransformer()
    ht.detect_initial_config(customers)

    # credit_card and id are pii and text columns
    ht.update_sdtypes({
        'credit_card': 'pii',
        'id': 'text'
    })

    ht.update_transformers({
        'credit_card': AnonymizedFaker(),
        'id': RegexGenerator(regex_format='id_[a-z]')
    })

    # Run
    ht.fit(customers)
    transformed = ht.transform(customers)
    reverse_transformed = ht.reverse_transform_subset(transformed[['last_login']])

    # Assert
    expected_transformed_columns = [
        'last_login',
        'email_optin',
        'age',
        'dollars_spent'
    ]
    assert all(expected_transformed_columns == transformed.columns)
    assert reverse_transformed.columns == ['last_login']


def test_set_config_with_supported_sdtypes():
    """Test when setting config with an sdtype that is supported by the transformer."""
    # Setup
    config = {
        'transformers': {
            'boolean_col': FrequencyEncoder(add_noise=True),
        },
        'sdtypes': {
            'boolean_col': 'boolean'
        }
    }
    ht = HyperTransformer()

    # Run and Assert
    ht.set_config(config)


def test_hyper_transformer_chained_transformers():
    """Test when a transformer is chained to another transformer.

    When the specified transformer indicates a next transformer, they should each be applied in
    order during the transform step, and then reversed during the reverse_transform.
    """
    # Setup
    class DoublingTransformer(BaseTransformer):
        INPUT_SDTYPE = 'numerical'

        def _fit(self, data):
            ...

        def _transform(self, data):
            return data * 2

        def _reverse_transform(self, data):
            return data / 2

    transformer3 = DoublingTransformer()
    transformer2 = DoublingTransformer()
    transformer2.output_properties[None]['next_transformer'] = transformer3
    transformer1 = DoublingTransformer()
    transformer1.output_properties[None]['next_transformer'] = transformer2

    ht = HyperTransformer()
    data = pd.DataFrame({'col': [1., 2, -1, 3, 1]})

    # Run and Assert
    ht.set_config({
        'sdtypes': {'col': 'numerical'},
        'transformers': {'col': transformer1}
    })
    ht.fit(data)

    transformed = ht.transform(data)
    expected_transform = pd.DataFrame({'col': [8., 16, -8, 24, 8]})
    pd.testing.assert_frame_equal(transformed, expected_transform)

    reverse_transformed = ht.reverse_transform(transformed)
    pd.testing.assert_frame_equal(reverse_transformed, data)


def test_hyper_transformer_chained_transformers_various_transformers():
    """Test when a transformer is chained to another transformer.

    When the specified transformer indicates a next transformer, they should each be applied in
    order during the transform step, and then reversed during the reverse_transform.
    """
    # Setup
    class AB(BaseTransformer):
        INPUT_SDTYPE = 'categorical'

        def _fit(self, data):
            self.output_properties = {
                None: {'sdtype': 'categorical', 'next_transformer': None},
                'a': {'sdtype': 'categorical', 'next_transformer': CD()},
                'b': {'sdtype': 'categorical', 'next_transformer': None},
            }

        def _transform(self, data):
            new_data = pd.DataFrame({f'{self.column_prefix}': data})
            new_data[f'{self.column_prefix}.a'] = data + 'a'
            new_data[f'{self.column_prefix}.b'] = data + 'b'
            return new_data

        def _reverse_transform(self, data):
            new_data = pd.DataFrame()
            new_data[f'{self.column_prefix}'] = data[f'{self.column_prefix}.a'].str[:-1]
            return new_data

    class CD(BaseTransformer):
        INPUT_SDTYPE = 'categorical'

        def _fit(self, data):
            self.output_properties = {
                'c': {'sdtype': 'categorical', 'next_transformer': None},
                'd': {'sdtype': 'categorical', 'next_transformer': E()}
            }

        def _transform(self, data):
            new_data = pd.DataFrame()
            new_data[f'{self.column_prefix}.c'] = data + 'c'
            new_data[f'{self.column_prefix}.d'] = data + 'd'
            return new_data

        def _reverse_transform(self, data):
            new_data = pd.DataFrame()
            new_data[f'{self.column_prefix}'] = data[f'{self.column_prefix}.c'].str[:-1]
            return new_data

    class E(BaseTransformer):
        INPUT_SDTYPE = 'categorical'

        def _fit(self, data):
            self.output_properties = {
                None: {'sdtype': 'categorical', 'next_transformer': None},
                'e': {'sdtype': 'categorical', 'next_transformer': None}
            }

        def _transform(self, data):
            new_data = pd.DataFrame({f'{self.column_prefix}': data})
            new_data[f'{self.column_prefix}.e'] = data + 'e'
            return new_data

        def _reverse_transform(self, data):
            new_data = pd.DataFrame()
            new_data[f'{self.column_prefix}'] = data[f'{self.column_prefix}.e'].str[:-1]
            return new_data

    ht = HyperTransformer()
    data = pd.DataFrame({
        'col': ['a', 'b', 'c'],
        'col.a': ['1', '2', '3'],
        'col#': ['_', '_', '_']
    })

    # Run and Assert
    ht.set_config({
        'sdtypes': {'col': 'categorical', 'col.a': 'categorical', 'col#': 'categorical'},
        'transformers': {'col': AB(), 'col.a': AB(), 'col#': E()}
    })
    ht.fit(data)
    transformed = ht.transform(data)
    expected = pd.DataFrame({
        'col##': ['a', 'b', 'c'],
        'col##.a.c': ['aac', 'bac', 'cac'],
        'col##.a.d': ['aad', 'bad', 'cad'],
        'col##.a.d.e': ['aade', 'bade', 'cade'],
        'col##.b': ['ab', 'bb', 'cb'],
        'col.a': ['1', '2', '3'],
        'col.a.a.c': ['1ac', '2ac', '3ac'],
        'col.a.a.d': ['1ad', '2ad', '3ad'],
        'col.a.a.d.e': ['1ade', '2ade', '3ade'],
        'col.a.b': ['1b', '2b', '3b'],
        'col#': ['_', '_', '_'],
        'col#.e': ['_e', '_e', '_e'],
    })
    pd.testing.assert_frame_equal(transformed, expected)

    reverse_transformed = ht.reverse_transform(transformed)
    pd.testing.assert_frame_equal(reverse_transformed, data)


def test_hyper_transformer_field_transformer_correctly_set():
    """Test field_transformers is correctly set through various methods."""
    # Setup
    ht = HyperTransformer()
    data = pd.DataFrame({'col': ['a', 'b', 'c']})

    # getting detected transformers should give the actual transformer
    ht.detect_initial_config(data)
    transformer = ht.get_config()['transformers']['col']
    transformer.new_attribute = 'abc'
    assert ht.get_config()['transformers']['col'].new_attribute == 'abc'

    # the actual transformer should be fitted
    ht.fit(data)
    transformer.new_attribute2 = '123'
    assert ht.get_config()['transformers']['col'].new_attribute == 'abc'
    assert ht.get_config()['transformers']['col'].new_attribute2 == '123'

    # if a transformer was set, it should use the provided instance
    fe = FrequencyEncoder()
    ht.set_config({'sdtypes': {'col': 'categorical'}, 'transformers': {'col': fe}})
    ht.fit(data)
    transformer = ht.get_config()['transformers']['col']
    assert transformer is fe

    # the three cases below make sure any form of acess to the field_transformers
    # correctly accesses and stores the actual transformers
    fe = FrequencyEncoder()
    ht.update_transformers({'col': fe})
    ht.fit(data)
    transformer = ht.get_config()['transformers']['col']
    assert transformer is fe

    ht.update_transformers_by_sdtype('categorical', transformer_name='FrequencyEncoder')
    transformer = ht.get_config()['transformers']['col']
    transformer.new_attribute3 = 'abc'
    ht.fit(data)
    assert ht.get_config()['transformers']['col'].new_attribute3 == 'abc'

    ht.update_sdtypes({'col': 'text'})
    transformer = ht.get_config()['transformers']['col']
    transformer.new_attribute3 = 'abc'
    ht.fit(data)
    assert ht.get_config()['transformers']['col'].new_attribute3 == 'abc'


def _get_hyper_transformer_with_random_transformers(data):
    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.update_sdtypes({
        'credit_card': 'pii',
        'name': 'text',
        'signup_day': 'datetime'
    })
    ht.update_transformers({
        'credit_card': AnonymizedFaker('credit_card', 'credit_card_number'),
        'balance': ClusterBasedNormalizer(max_clusters=3),
        'name': RegexGenerator()
    })
    ht.update_transformers_by_sdtype(
        'categorical',
        transformer_name='FrequencyEncoder',
        transformer_parameters={'add_noise': True}
    )

    return ht


def test_hyper_transformer_reset_randomization():
    """Test that the random seeds are properly set and reset.

    If two separate ``HyperTransformer``s are fit, they should have the same parameters
    and produce the same data when transforming. Successive transforming calls should
    yield different results.

    For reverse transforming, different values should be seen if randomization is involved
    unless ``reset_randomization`` is called.
    """
    # Setup
    data = pd.DataFrame({
        'credit_card': ['123456789', '987654321', '192837645', '918273465', '123789456'],
        'age': [18, 25, 54, 60, 31],
        'name': ['Bob', 'Jane', 'Jack', 'Jill', 'Joe'],
        'signup_day': ['1/1/2020', np.nan, '4/1/2019', '12/1/2008', '5/16/2016'],
        'balance': [250, 5400, 150000, np.nan, 91000],
        'card_type': ['Visa', 'Visa', 'Master Card', 'Amex', 'Visa']
    })
    ht1 = _get_hyper_transformer_with_random_transformers(data)
    ht2 = _get_hyper_transformer_with_random_transformers(data)

    # Test transforming multiple times with different transformers
    expected_first_transformed = pd.DataFrame({
        'age': [18.0, 25.0, 54.0, 60.0, 31.0],
        'signup_day': [1.577837e+18, 1.455840e+18, 1.554077e+18, 1.228090e+18, 1.463357e+18],
        'balance.normalized': [
            -2.693016e-01,
            -2.467182e-01,
            3.873711e-01,
            9.571797e-17,
            1.286486e-01
        ],
        'balance.component': [0.0, 0, 0, 0, 0],
        'card_type': [
            0.31440326318001877,
            0.2879792449993428,
            0.7147347796329043,
            0.9397813187279729,
            0.25144169889394075
        ]
    })
    expected_second_transformed = pd.DataFrame({
        'age': [18.0, 25.0, 54.0, 60.0, 31.0],
        'signup_day': [1.577837e+18, 1.455840e+18, 1.554077e+18, 1.228090e+18, 1.463357e+18],
        'balance.normalized': [
            -2.693016e-01,
            -2.467182e-01,
            3.873711e-01,
            9.571797e-17,
            1.286486e-01
        ],
        'balance.component': [0.0, 0, 0, 0, 0],
        'card_type': [
            0.20248666820805558,
            0.4408301080724041,
            0.7082705433167992,
            0.8911691002937682,
            0.2679993642397502
        ]
    })

    ht1.fit(data)
    first_transformed1 = ht1.transform(data)
    ht2.fit(data)
    first_transformed2 = ht2.transform(data)
    second_transformed1 = ht1.transform(data)

    pd.testing.assert_frame_equal(first_transformed1, expected_first_transformed)
    pd.testing.assert_frame_equal(first_transformed2, expected_first_transformed)
    pd.testing.assert_frame_equal(second_transformed1, expected_second_transformed)

    # test reverse transforming multiple times with different tranformers
    expected_first_reverse = pd.DataFrame({
        'credit_card': [
            '3564483245479407',
            '4863061245886958069',
            '4039466324278480',
            '4217004814656859',
            '4343691397776091'
        ],
        'age': [18, 25, 54, 60, 31],
        'name': ['AAAAA', 'AAAAB', 'AAAAC', 'AAAAD', 'AAAAE'],
        'signup_day': ['01/01/2020', '02/19/2016', '04/01/2019', np.nan, np.nan],
        'balance': [250, 5400, 150000, 61662.5, 91000],
        'card_type': ['Visa', 'Visa', 'Master Card', 'Amex', 'Visa']
    })
    expected_second_reverse = pd.DataFrame({
        'credit_card': [
            '4208002654643',
            '3547584322792794',
            '30187802217181',
            '4138954513622487900',
            '346502559595986'
        ],
        'age': [18, 25, 54, 60, 31],
        'name': ['AAAAF', 'AAAAG', 'AAAAH', 'AAAAI', 'AAAAJ'],
        'signup_day': ['01/01/2020', np.nan, '04/01/2019', '12/01/2008', np.nan],
        'balance': [np.nan, 5400, np.nan, 61662.5, 91000],
        'card_type': ['Visa', 'Visa', 'Master Card', 'Amex', 'Visa']
    })
    first_reverse1 = ht1.reverse_transform(first_transformed1)
    first_reverse2 = ht2.reverse_transform(first_transformed1)
    second_reverse1 = ht1.reverse_transform(first_transformed1)
    pd.testing.assert_frame_equal(first_reverse1, expected_first_reverse)
    pd.testing.assert_frame_equal(first_reverse2, expected_first_reverse)
    pd.testing.assert_frame_equal(expected_second_reverse, second_reverse1)

    # Test resetting randomization
    ht1.reset_randomization()

    transformed_post_reset = ht1.reverse_transform(first_transformed1)
    pd.testing.assert_frame_equal(transformed_post_reset, expected_first_reverse)


def test_hyper_transformer_cluster_based_normalizer_randomization():
    """Test that the ``ClusterBasedNormalizer`` handles randomization correctly.

    If the ``ClusterBasedNormalizer`` transforms the same data multiple times,
    it may yield different results due to randomness. However, if a new instance is created,
    each matching call should yield the same results (ie. call 1 of the first transformer
    should match call 1 of the second).
    """
    data = get_demo(100)
    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.update_transformers({
        'age': ClusterBasedNormalizer()
    })
    ht.fit(data)
    transformed1 = ht.transform(data)
    transformed2 = ht.transform(data)

    assert any(transformed1['age.normalized'] != transformed2['age.normalized'])

    ht2 = HyperTransformer()
    ht2.detect_initial_config(data)
    ht2.update_transformers({
        'age': ClusterBasedNormalizer()
    })
    ht2.fit(data)

    pd.testing.assert_frame_equal(transformed1, ht2.transform(data))


def test_hypertransformer_anonymized_faker():
    """Test ``AnonymizedFaker`` generates different random values for different columns.

    Issue: https://github.com/sdv-dev/RDT/issues/619.
    """
    # Setup
    data = pd.DataFrame({
        'id1': ['a', 'b', 'c'],
        'id2': ['d', 'e', 'f'],
    })
    ht = HyperTransformer()

    # Run - simple run
    ht.detect_initial_config(data)
    ht.update_sdtypes({
        'id1': 'pii',
        'id2': 'pii'
    })
    ht.update_transformers({
        'id1': AnonymizedFaker(),
        'id2': AnonymizedFaker()
    })
    ht.fit(data)
    transformed = ht.transform(data)
    reverse_transformed1 = ht.reverse_transform(transformed)

    # Assert
    assert reverse_transformed1['id1'].tolist() != reverse_transformed1['id2'].tolist()

    # Run - make sure transforming again returns different values than the original transform
    transformed = ht.transform(data)
    reverse_transformed2 = ht.reverse_transform(transformed)

    # Assert
    assert reverse_transformed2['id1'].tolist() != reverse_transformed2['id2'].tolist()
    assert reverse_transformed1['id1'].tolist() != reverse_transformed2['id1'].tolist()
    assert reverse_transformed1['id2'].tolist() != reverse_transformed2['id2'].tolist()

    # Run - make sure resetting randomization works
    ht.reset_randomization()
    transformed = ht.transform(data)
    reverse_transformed3 = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(reverse_transformed1, reverse_transformed3)


def test_hypertransformer_pseudo_anonymized_faker():
    """Test ``PseudoAnonymizedFaker`` generates different random values for different columns."""
    # Setup
    data = pd.DataFrame({
        'id1': ['a', 'b', 'c'],
        'id2': ['a', 'b', 'c'],
    })
    ht = HyperTransformer()

    # Run
    ht.detect_initial_config(data)
    ht.update_sdtypes({
        'id1': 'pii',
        'id2': 'pii'
    })
    ht.update_transformers({
        'id1': PseudoAnonymizedFaker(),
        'id2': PseudoAnonymizedFaker()
    })
    ht.fit(data)
    transformed = ht.transform(data)
    reverse_transformed1 = ht.reverse_transform(transformed)

    # Assert
    assert reverse_transformed1['id1'].tolist() != reverse_transformed1['id2'].tolist()

    # Run - run it again on the exact same data
    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.update_sdtypes({
        'id1': 'pii',
        'id2': 'pii'
    })
    ht.update_transformers({
        'id1': PseudoAnonymizedFaker(),
        'id2': PseudoAnonymizedFaker()
    })
    ht.fit(data)
    transformed = ht.transform(data)
    reverse_transformed2 = ht.reverse_transform(transformed)

    # Assert - different instances of the same transformer should return the same result
    assert reverse_transformed1['id1'].tolist() == reverse_transformed2['id1'].tolist()


def test_hypertransformer_anonymized_faker_multi_table():
    """Test ``AnonymizedFaker`` generates different values for columns with same name."""
    # Setup
    data1 = pd.DataFrame({
        'id1': ['a', 'b', 'c'],
        'id2': ['a', 'b', 'c'],
    })
    data2 = pd.DataFrame({
        'id1': ['d', 'e', 'f'],
        'id2': ['d', 'e', 'f'],
    })
    ht = HyperTransformer()

    # Run on data1
    ht.detect_initial_config(data1)
    ht.update_sdtypes({
        'id1': 'pii',
        'id2': 'pii'
    })
    ht.update_transformers({
        'id1': AnonymizedFaker(),
        'id2': PseudoAnonymizedFaker()
    })
    ht.fit(data1)
    transformed = ht.transform(data1)
    reverse_transformed1 = ht.reverse_transform(transformed)

    # Run on data2
    ht.detect_initial_config(data2)
    ht.update_sdtypes({
        'id1': 'pii',
        'id2': 'pii'
    })
    ht.update_transformers({
        'id1': AnonymizedFaker(),
        'id2': PseudoAnonymizedFaker()
    })
    ht.fit(data2)
    transformed = ht.transform(data2)
    reverse_transformed2 = ht.reverse_transform(transformed)

    # Assert
    assert reverse_transformed1['id1'].tolist() != reverse_transformed2['id1'].tolist()
    assert reverse_transformed1['id2'].tolist() != reverse_transformed2['id2'].tolist()


def test_random_seed():
    # Setup
    data = pd.DataFrame({
        'num1': [1, np.nan, 2] * 10,
        'num2': [1, np.nan, 2] * 10,
        'num3': [1, np.nan, 2] * 10,
        'num4': [1, np.nan, 2] * 10,
        'num5': [1, np.nan, 2] * 10,
        'num6': [1, np.nan, 2] * 10,
        'date1': [np.datetime64('2020-10-10'), np.datetime64('2021-11-11'), np.nan] * 10,
        'date2': [np.datetime64('2020-10-10'), np.datetime64('2021-11-11'), np.nan] * 10,
        'date3': [np.datetime64('2020-10-10'), np.datetime64('2021-11-11'), np.nan] * 10,
        'date4': [np.datetime64('2020-10-10'), np.datetime64('2021-11-11'), np.nan] * 10,
    })

    ht = HyperTransformer()
    ht.detect_initial_config(data)
    ht.update_transformers({
        'num1': FloatFormatter(),
        'num2': FloatFormatter(),
        'num3': ClusterBasedNormalizer(),
        'num4': ClusterBasedNormalizer(),
        'num5': GaussianNormalizer(),
        'num6': GaussianNormalizer(),
        'date1': UnixTimestampEncoder(),
        'date2': UnixTimestampEncoder(),
        'date3': OptimizedTimestampEncoder(),
        'date4': OptimizedTimestampEncoder(),
    })

    # Run
    ht.fit(data)
    transformed = ht.transform(data)
    reversed1 = ht.reverse_transform(transformed)

    # Assert
    assert reversed1['num1'].isna().tolist() != reversed1['num2'].isna().tolist()
    assert reversed1['num3'].isna().tolist() != reversed1['num4'].isna().tolist()
    assert reversed1['num5'].isna().tolist() != reversed1['num6'].isna().tolist()
    assert reversed1['date1'].isna().tolist() != reversed1['date2'].isna().tolist()
    assert reversed1['date3'].isna().tolist() != reversed1['date4'].isna().tolist()

    # Run
    ht.reset_randomization()
    transformed = ht.transform(data)
    reversed2 = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(reversed1, reversed2)



import re

import numpy as np
import pandas as pd
import pytest

from rdt.errors import TransformerProcessingError
from rdt.transformers.pii import AnonymizedFaker, PseudoAnonymizedFaker


def test_anonymizedfaker():
    """End to end test with the default settings of the ``AnonymizedFaker``."""
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e']
    })

    instance = AnonymizedFaker()
    transformed = instance.fit_transform(data, 'username')

    reverse_transform = instance.reverse_transform(transformed)
    expected_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5]
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    assert len(reverse_transform['username']) == 5


def test_anonymizedfaker_get_supported_sdtypes():
    """Test that the correct supported sdtypes are returned."""
    # Run
    supported_sdtypes = AnonymizedFaker.get_supported_sdtypes()

    # Assert
    assert supported_sdtypes == ['pii']


def test_anonymizedfaker_custom_provider():
    """End to end test with a custom provider and function for the ``AnonymizedFaker``."""
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e'],
        'cc': [
            '2276346007210438',
            '4149498289355',
            '213144860944676',
            '4514775286178',
            '213133122335401'
        ]
    })

    instance = AnonymizedFaker('credit_card', 'credit_card_number')
    transformed = instance.fit_transform(data, 'cc')
    reverse_transform = instance.reverse_transform(transformed)

    expected_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e'],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    assert len(reverse_transform['cc']) == 5


def test_anonymizedfaker_with_nans():
    """End to end test with the default settings of the ``AnonymizedFaker`` with ``nan`` values."""
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', np.nan, 'c', 'd', 'e']
    })

    instance = AnonymizedFaker()
    transformed = instance.fit_transform(data, 'username')
    reverse_transform = instance.reverse_transform(transformed)

    expected_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    assert len(reverse_transform['username']) == 5
    assert reverse_transform['username'].isna().sum() == 1


def test_anonymizedfaker_with_nans_missing_value_generation_none():
    """End to end test settings missing_value_generation=None."""
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', np.nan, 'c', 'd', 'e']
    })

    instance = AnonymizedFaker(missing_value_generation=None)
    transformed = instance.fit_transform(data, 'username')
    reverse_transform = instance.reverse_transform(transformed)

    expected_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    assert len(reverse_transform['username']) == 5
    assert reverse_transform['username'].isna().sum() == 0


def test_anonymizedfaker_custom_provider_with_nans():
    """End to end test with a custom provider for the ``AnonymizedFaker`` with `` nans``."""
    data = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e'],
        'cc': [
            '2276346007210438',
            np.nan,
            '213144860944676',
            '4514775286178',
            '213133122335401'
        ]
    })

    instance = AnonymizedFaker(
        'credit_card',
        'credit_card_number',
    )
    transformed = instance.fit_transform(data, 'cc')
    reverse_transform = instance.reverse_transform(transformed)

    expected_transformed = pd.DataFrame({
        'id': [1, 2, 3, 4, 5],
        'username': ['a', 'b', 'c', 'd', 'e'],
    })

    pd.testing.assert_frame_equal(transformed, expected_transformed)
    assert len(reverse_transform['cc']) == 5
    assert reverse_transform['cc'].isna().sum() == 1


def test_anonymizedfaker_enforce_uniqueness():
    """Test that ``AnonymizedFaker`` works with uniqueness.

    Also ensure that when we call ``reset_randomization`` the generator will be able to
    create values again.
    """
    data = pd.DataFrame({
        'job': np.arange(500)
    })

    instance = AnonymizedFaker('job', 'job', enforce_uniqueness=True)
    transformed = instance.fit_transform(data, 'job')
    reverse_transform = instance.reverse_transform(transformed)

    # Assert
    assert len(reverse_transform['job'].unique()) == 500

    error_msg = re.escape(
        'The Faker function you specified is not able to generate 500 unique '
        'values. Please use a different Faker function for column '
        "('job')."
    )
    with pytest.raises(TransformerProcessingError, match=error_msg):
        instance.reverse_transform(transformed)

    instance.reset_randomization()
    instance.reverse_transform(transformed)


def test_pseudoanonymizedfaker():
    """End to end test with the default settings of the ``PseudoAnonymizedFaker``."""
    data = pd.DataFrame({
        'animals': ['cat', 'dog', 'parrot', 'monkey']
    })

    instance = PseudoAnonymizedFaker()

    transformed = instance.fit_transform(data, 'animals')
    reverse_transformed = instance.reverse_transform(transformed)

    assert transformed.columns == ['animals']
    pd.testing.assert_series_equal(
        reverse_transformed['animals'].map(instance._reverse_mapping_dict),
        data['animals']
    )
    assert set(reverse_transformed['animals']).intersection(set(instance._mapping_dict)) == set()
    assert len(reverse_transformed) == len(transformed) == 4


def test_pseudoanonymizedfaker_with_nans():
    """End to end test with the default settings of the ``PseudoAnonymizedFaker`` and ``nans``."""
    data = pd.DataFrame({
        'animals': ['cat', 'dog', np.nan, 'monkey']
    })

    instance = PseudoAnonymizedFaker()

    transformed = instance.fit_transform(data, 'animals')
    reverse_transformed = instance.reverse_transform(transformed)

    assert transformed.columns == ['animals']
    pd.testing.assert_series_equal(
        reverse_transformed['animals'].map(instance._reverse_mapping_dict),
        data['animals']
    )
    assert set(reverse_transformed['animals']).intersection(set(instance._mapping_dict)) == set()
    assert len(reverse_transformed) == len(transformed) == 4


def test_pseudoanonymizedfaker_with_custom_provider():
    """End to end test with custom settings of the ``PseudoAnonymizedFaker``."""
    data = pd.DataFrame({
        'animals': ['cat', 'dog', np.nan, 'monkey']
    })

    instance = PseudoAnonymizedFaker('credit_card', 'credit_card_number')

    transformed = instance.fit_transform(data, 'animals')
    reverse_transformed = instance.reverse_transform(transformed)

    assert transformed.columns == ['animals']
    pd.testing.assert_series_equal(
        reverse_transformed['animals'].map(instance._reverse_mapping_dict),
        data['animals']
    )
    assert set(reverse_transformed['animals']).intersection(set(instance._mapping_dict)) == set()
    assert len(reverse_transformed) == len(transformed) == 4



import numpy as np
import pandas as pd

from rdt.transformers import BinaryEncoder


class TestBinaryEncoder:

    def test_boolean_some_nans(self):
        """Test BinaryEncoder on input with some nan values.

        Ensure that the BinaryEncoder can fit, transform, and reverse transform on boolean data
        with Nones. Expect that the reverse transformed data is the same as the input, but None
        becomes nan and the False/nan values can be interchanged.

        Also ensures that the intermediate transformed data is unchanged after reversing.

        Input:
            - boolean data with None values
        Output:
            - The reversed transformed data
        """
        # Setup
        data = pd.DataFrame([True, False, None, False], columns=['bool'])
        column = 'bool'
        transformer = BinaryEncoder()

        # Run
        transformer.fit(data, column)
        transformed = transformer.transform(data)
        unchanged_transformed = transformed.copy()
        reverse = transformer.reverse_transform(transformed)

        # Assert
        np.testing.assert_array_equal(unchanged_transformed, transformed)
        assert reverse['bool'][0] in {True, np.nan}
        for value in reverse['bool'][1:]:
            assert value is False or np.isnan(value)

    def test_boolean_missing_value_replacement_mode(self):
        """Test BinaryEncoder when `missing_value_replacement` is set to 'mode'.

        Ensure that the BinaryEncoder can fit, transform, and reverse transform on
        boolean data when `missing_value_replacement` is set to `'mode'` and
        `missing_value_generation` is set to 'from_column'. Expect that the reverse
        transformed data is the same as the input.
        """
        # Setup
        data = pd.DataFrame([True, True, None, False], columns=['bool'])
        column = 'bool'
        transformer = BinaryEncoder(
            missing_value_replacement='mode',
            missing_value_generation='from_column'
        )

        # Run
        transformer.fit(data, column)
        transformed = transformer.transform(data)
        reverse = transformer.reverse_transform(transformed)

        # Assert
        expected_transformed = pd.DataFrame({
            'bool': [1., 1., 1., 0.],
            'bool.is_null': [0., 0., 1., 0.]
        })
        pd.testing.assert_frame_equal(transformed, expected_transformed)
        pd.testing.assert_frame_equal(reverse, data)

    def test_boolean_missing_value_generation_none(self):
        """Test the BinaryEncoder when ``missing_value_generation`` is None.

        In this test, the nans should be replacd by the mode on the transformed data.
        """
        # Setup
        data = pd.DataFrame([True, True, None, False], columns=['bool'])
        column = 'bool'
        transformer = BinaryEncoder(
            missing_value_replacement='mode',
            missing_value_generation=None
        )

        # Run
        transformer.fit(data, column)
        transformed = transformer.transform(data)
        reverse = transformer.reverse_transform(transformed)

        # Assert
        expected_transformed = pd.DataFrame({'bool': [1., 1., 1., 0.]})
        expected_reversed = pd.DataFrame({'bool': [True, True, True, False]})
        pd.testing.assert_frame_equal(transformed, expected_transformed)
        pd.testing.assert_frame_equal(reverse, expected_reversed, check_dtype=False)




import sys
from types import ModuleType
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest

import rdt
from rdt import _find_addons, get_demo


@pytest.fixture()
def mock_rdt():
    rdt_module = sys.modules['rdt']
    rdt_mock = Mock()
    rdt_mock.submodule.__name__ = 'rdt.submodule'
    sys.modules['rdt'] = rdt_mock
    yield rdt_mock
    sys.modules['rdt'] = rdt_module


def test_get_demo():
    demo = get_demo()

    assert list(demo.columns) == [
        'last_login', 'email_optin', 'credit_card', 'age', 'dollars_spent'
    ]
    assert len(demo) == 5
    assert list(demo.isna().sum(axis=0)) == [1, 1, 1, 0, 1]


def test_get_demo_many_rows():
    demo = get_demo(10)

    login_dates = pd.Series([
        '2021-06-26', '2021-02-10', 'NaT', '2020-09-26', '2020-12-22', '2019-11-27',
        '2002-05-10', '2014-10-04', '2014-03-19', '2015-09-13'
    ], dtype='datetime64[ns]')
    email_optin = [False, False, False, True, np.nan, np.nan, False, True, False, False]
    credit_card = [
        'VISA', 'VISA', 'AMEX', np.nan, 'DISCOVER', 'AMEX', 'AMEX', 'DISCOVER', 'DISCOVER', 'VISA'
    ]
    age = [29, 18, 21, 45, 32, 50, 93, 75, 39, 66]
    dollars_spent = [99.99, np.nan, 2.50, 25.00, 19.99, 52.48, 39.99, 4.67, np.nan, 23.28]

    expected = pd.DataFrame({
        'last_login': login_dates,
        'email_optin': email_optin,
        'credit_card': credit_card,
        'age': age,
        'dollars_spent': dollars_spent
    })

    pd.testing.assert_frame_equal(demo, expected)


@patch.object(rdt, 'iter_entry_points')
def test__find_addons_module(entry_points_mock, mock_rdt):
    """Test loading an add-on."""
    # Setup
    add_on_mock = Mock(spec=ModuleType)
    entry_point = Mock()
    entry_point.name = 'rdt.submodule.entry_name'
    entry_point.load.return_value = add_on_mock
    entry_points_mock.return_value = [entry_point]

    # Run
    _find_addons()

    # Assert
    entry_points_mock.assert_called_once_with(group='rdt_modules')
    assert mock_rdt.submodule.entry_name == add_on_mock
    assert sys.modules['rdt.submodule.entry_name'] == add_on_mock


@patch.object(rdt, 'iter_entry_points')
def test__find_addons_object(entry_points_mock, mock_rdt):
    """Test loading an add-on."""
    # Setup
    entry_point = Mock()
    entry_point.name = 'rdt.submodule:entry_object.entry_method'
    entry_point.load.return_value = 'new_method'
    entry_points_mock.return_value = [entry_point]

    # Run
    _find_addons()

    # Assert
    entry_points_mock.assert_called_once_with(group='rdt_modules')
    assert mock_rdt.submodule.entry_object.entry_method == 'new_method'


@patch('warnings.warn')
@patch('rdt.iter_entry_points')
def test__find_addons_bad_addon(entry_points_mock, warning_mock):
    """Test failing to load an add-on generates a warning."""
    # Setup
    def entry_point_error():
        raise ValueError()

    bad_entry_point = Mock()
    bad_entry_point.name = 'bad_entry_point'
    bad_entry_point.module_name = 'bad_module'
    bad_entry_point.load.side_effect = entry_point_error
    entry_points_mock.return_value = [bad_entry_point]
    msg = 'Failed to load "bad_entry_point" from "bad_module".'

    # Run
    _find_addons()

    # Assert
    entry_points_mock.assert_called_once_with(group='rdt_modules')
    warning_mock.assert_called_once_with(msg)


@patch('warnings.warn')
@patch('rdt.iter_entry_points')
def test__find_addons_wrong_base(entry_points_mock, warning_mock):
    """Test incorrect add-on name generates a warning."""
    # Setup
    bad_entry_point = Mock()
    bad_entry_point.name = 'bad_base.bad_entry_point'
    entry_points_mock.return_value = [bad_entry_point]
    msg = (
        "Failed to set 'bad_base.bad_entry_point': expected base module to be 'rdt', found "
        "'bad_base'."
    )

    # Run
    _find_addons()

    # Assert
    entry_points_mock.assert_called_once_with(group='rdt_modules')
    warning_mock.assert_called_once_with(msg)


@patch('warnings.warn')
@patch('rdt.iter_entry_points')
def test__find_addons_missing_submodule(entry_points_mock, warning_mock):
    """Test incorrect add-on name generates a warning."""
    # Setup
    bad_entry_point = Mock()
    bad_entry_point.name = 'rdt.missing_submodule.new_submodule'
    entry_points_mock.return_value = [bad_entry_point]
    msg = (
        "Failed to set 'rdt.missing_submodule.new_submodule': module 'rdt' has no attribute "
        "'missing_submodule'."
    )

    # Run
    _find_addons()

    # Assert
    entry_points_mock.assert_called_once_with(group='rdt_modules')
    warning_mock.assert_called_once_with(msg)


@patch('warnings.warn')
@patch('rdt.iter_entry_points')
def test__find_addons_module_and_object(entry_points_mock, warning_mock):
    """Test incorrect add-on name generates a warning."""
    # Setup
    bad_entry_point = Mock()
    bad_entry_point.name = 'rdt.missing_submodule:new_object'
    entry_points_mock.return_value = [bad_entry_point]
    msg = (
        "Failed to set 'rdt.missing_submodule:new_object': cannot add 'new_object' to unknown "
        "submodule 'rdt.missing_submodule'."
    )

    # Run
    _find_addons()

    # Assert
    entry_points_mock.assert_called_once_with(group='rdt_modules')
    warning_mock.assert_called_once_with(msg)


@patch('warnings.warn')
@patch.object(rdt, 'iter_entry_points')
def test__find_addons_missing_object(entry_points_mock, warning_mock, mock_rdt):
    """Test incorrect add-on name generates a warning."""
    # Setup
    bad_entry_point = Mock()
    bad_entry_point.name = 'rdt.submodule:missing_object.new_method'
    entry_points_mock.return_value = [bad_entry_point]
    msg = ("Failed to set 'rdt.submodule:missing_object.new_method': missing_object.")

    del mock_rdt.submodule.missing_object

    # Run
    _find_addons()

    # Assert
    entry_points_mock.assert_called_once_with(group='rdt_modules')
    warning_mock.assert_called_once_with(msg)



"""Unit tests for the NullTransformer."""

import re
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from rdt.errors import TransformerInputError
from rdt.transformers import NullTransformer


class TestNullTransformer:

    def test___init__default(self):
        """Test the initialization without passing any default arguments.

        When no arguments are passed, the attributes should be populated
        with the right values.
        """
        # Run
        transformer = NullTransformer()

        # Assert
        assert transformer._missing_value_replacement is None
        assert transformer._missing_value_generation == 'random'

    def test___init__not_default(self):
        """Test the initialization passing values different than defaults.

        When arguments are passed, the attributes should be populated
        with the right values.

        Input:
            - Values different than the defaults.

        Expected Side Effects:
            - The attributes should be populated with the given values.
        """
        # Run
        transformer = NullTransformer('a_missing_value_replacement', None)

        # Assert
        assert transformer._missing_value_replacement == 'a_missing_value_replacement'
        assert transformer._missing_value_generation is None

    def test___init__raises_error(self):
        """Test the initialization passing invalid values for ``missing_value_generation``."""
        # Setup
        error_msg = re.escape(
            "'missing_value_generation' must be one of the following values: None, 'from_column' "
            "or 'random'."
        )

        # Run / Assert
        with pytest.raises(TransformerInputError, match=error_msg):
            NullTransformer('mean', 'None')

    def test_models_missing_values(self):
        """Test the models_missing_values method.

        Test that when ``missing_value_generation`` is ``'from_column'``, this returns
        ``True``.
        """
        # Setup
        transformer = NullTransformer('something', missing_value_generation='from_column')

        # Run
        models_missing_values = transformer.models_missing_values()

        # Assert
        assert models_missing_values is True

    def test_models_missing_values_missing_value_generation_is_none(self):
        """Test the models_missing_values method.

        Test that when ``missing_value_generation`` is other than ``'from_column'``, this returns
        ``False``.
        """
        # Setup
        none_transformer = NullTransformer('something', missing_value_generation=None)
        random_transformer = NullTransformer('something', missing_value_generation='random')

        # Run
        none_models_missing_values = none_transformer.models_missing_values()
        random_models_missing_values = random_transformer.models_missing_values()

        # Assert
        assert none_models_missing_values is False
        assert random_models_missing_values is False

    def test__get_missing_value_replacement_scalar(self):
        """Test _get_missing_value_replacement when a scalar value is passed.

        If a missing_value_replacement different from None, 'mean' or 'mode' is
        passed to __init__, that value is returned.

        Setup:
            - NullTransformer passing a specific missing_value_replacement
              that is not None, mean or mode.

        Input:
            - A Series with some values.
            - A np.array with boolean values.

        Expected Output:
            - The value passed to __init__
        """
        # Setup
        transformer = NullTransformer('a_missing_value_replacement')

        # Run
        data = pd.Series([1, np.nan, 3], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 'a_missing_value_replacement'

    def test__get_missing_value_replacement_none_numerical(self):
        """Test _get_missing_value_replacement when missing_value_replacement is None.

        If the missing_value_replacement is None and the data is numerical,
        the output fill value should be the mean of the input data.

        Setup:
            - NullTransformer passing with default arguments.

        Input:
            - An Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The mean of the inputted Series.
        """
        # Setup
        transformer = NullTransformer('mean')

        # Run
        data = pd.Series([1, 2, np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 1.5

    def test__get_missing_value_replacement_none_not_numerical(self):
        """Test _get_missing_value_replacement when missing_value_replacement is None.

        If the missing_value_replacement is None and the data is not numerical,
        the output fill value should be the mode of the input data.

        Setup:
            - NullTransformer with default arguments.

        Input:
            - An Series filled with string values with variable frequency and
              at least one null value.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The most frequent value in the input series.
        """
        # Setup
        transformer = NullTransformer('mode')

        # Run
        data = pd.Series(['a', 'b', 'b', np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 'b'

    def test__get_missing_value_replacement_mean(self):
        """Test _get_missing_value_replacement when missing_value_replacement is mean.

        If the missing_value_replacement is mean the output fill value should be the
        mean of the input data.

        Setup:
            - NullTransformer passing 'mean' as the missing_value_replacement.

        Input:
            - A Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The mode of the inputted Series.
        """
        # Setup
        transformer = NullTransformer('mean')

        # Run
        data = pd.Series([1, 2, np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 1.5

    @patch('rdt.transformers.null.LOGGER')
    def test__get_missing_value_replacement_mean_only_nans(self, logger_mock):
        """Test when missing_value_replacement is mean and data only contains nans."""
        # Setup
        transformer = NullTransformer('mean')
        data = pd.Series([float('nan'), None, np.nan], name='abc')

        # Run
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        logger_mock.info.assert_called_once_with(
            "'missing_value_replacement' cannot be set to 'mean'"
            ' when the provided data only contains NaNs. Using 0 instead.'
        )
        assert missing_value_replacement == 0

    def test__get_missing_value_replacement_mode(self):
        """Test _get_missing_value_replacement when missing_value_replacement is 'mode'.

        If the missing_value_replacement is 'mode' the output fill value should be the
        mode of the input data.

        Setup:
            - NullTransformer passing 'mode' as the missing_value_replacement.

        Input:
            - A Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The most frequent value in the input series.
        """
        # Setup
        transformer = NullTransformer('mode')

        # Run
        data = pd.Series([1, 2, 2, np.nan], name='abc')
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        assert missing_value_replacement == 2

    @patch('rdt.transformers.null.LOGGER')
    def test__get_missing_value_replacement_mode_only_nans(self, logger_mock):
        """Test when missing_value_replacement is mode and data only contains nans."""
        # Setup
        transformer = NullTransformer('mode')
        data = pd.Series([float('nan'), None, np.nan], name='abc')

        # Run
        missing_value_replacement = transformer._get_missing_value_replacement(data)

        # Assert
        logger_mock.info.assert_called_once_with(
            "'missing_value_replacement' cannot be set to 'mode' when "
            'the provided data only contains NaNs. Using 0 instead.'
        )
        assert missing_value_replacement == 0

    def test_fit_missing_value_generation_is_none_and_nulls(self):
        """Test fit when ``missing_value_generation`` is ``None`` and there are nulls.

        Nothing has been learned and the nulls stay as ``None``. The ``_missing_value_replacement``
        should be set to the mean.
        """
        # Setup
        transformer = NullTransformer(
            missing_value_replacement='mean',
            missing_value_generation=None
        )

        # Run
        data = pd.Series([1, 2, np.nan])
        transformer.fit(data)

        # Assert
        assert transformer.nulls is None
        assert transformer._missing_value_replacement == 1.5

    def test_fit_missing_value_generation_from_column_and_no_nulls(self):
        """Test fit when ``missing_value_generation`` is 'from_column' and there are nulls."""
        # Setup
        transformer = NullTransformer(missing_value_generation='from_column')

        # Run
        data = pd.Series(['a', 'b', 'b'])
        transformer.fit(data)

        # Assert
        assert not transformer.nulls
        assert transformer._missing_value_generation is None
        assert transformer._missing_value_replacement is None

    def test_fit_with_multiple_missing_value_generations(self):
        """Test fit when ``missing_value_generation`` is set to its three possibilities.

        Test that when there are multiple scenarios given, the null transformer is able to
        either learn the mode or mean to replace the value but following the
        ``missing_value_generation`` strategy.

        Notice that this test covers 5 scenarios at once.
        """
        # Setup
        missing_value_generation_random_nulls = NullTransformer(
            missing_value_replacement='mode',
            missing_value_generation='random'
        )
        missing_value_generation_random_no_nulls = NullTransformer(
            missing_value_replacement='mode',
            missing_value_generation='random'
        )
        missing_value_generation_column_nulls = NullTransformer(
            missing_value_replacement='mean',
            missing_value_generation='from_column'
        )
        missing_value_generation_column_no_nulls = NullTransformer(
            missing_value_replacement='mean',
            missing_value_generation='from_column'
        )

        missing_value_generation_none_int = NullTransformer(
            missing_value_replacement='mean',
            missing_value_generation=None
        )
        missing_value_generation_none_str = NullTransformer(
            missing_value_replacement='mode',
            missing_value_generation=None
        )

        nulls_str = pd.Series(['a', 'b', 'b', np.nan])
        no_nulls_str = pd.Series(['a', 'b', 'b', 'c'])
        nulls_int = pd.Series([1, 2, 3, np.nan])
        no_nulls_int = pd.Series([1, 2, 3, 4])

        # Run
        missing_value_generation_random_nulls.fit(nulls_str)
        missing_value_generation_random_no_nulls.fit(no_nulls_str)
        missing_value_generation_column_nulls.fit(nulls_int)
        missing_value_generation_column_no_nulls.fit(no_nulls_int)
        missing_value_generation_none_int.fit(nulls_int)
        missing_value_generation_none_str.fit(nulls_str)

        # Assert
        assert missing_value_generation_random_nulls._missing_value_generation == 'random'
        assert missing_value_generation_random_nulls.nulls
        assert missing_value_generation_random_nulls._missing_value_replacement == 'b'

        assert missing_value_generation_random_no_nulls._missing_value_generation == 'random'
        assert not missing_value_generation_random_no_nulls.nulls
        assert missing_value_generation_random_no_nulls._missing_value_replacement == 'b'

        assert missing_value_generation_column_nulls._missing_value_generation == 'from_column'
        assert missing_value_generation_column_nulls.nulls
        assert missing_value_generation_column_nulls._missing_value_replacement == 2

        assert missing_value_generation_column_no_nulls._missing_value_generation is None
        assert not missing_value_generation_column_no_nulls.nulls
        assert missing_value_generation_column_no_nulls._missing_value_replacement == 2.5

        assert missing_value_generation_none_int._missing_value_generation is None
        assert missing_value_generation_none_int.nulls is None
        assert missing_value_generation_none_int._missing_value_replacement == 2

        assert missing_value_generation_none_str._missing_value_generation is None
        assert missing_value_generation_none_str.nulls is None
        assert missing_value_generation_none_str._missing_value_replacement == 'b'

    def test_transform__missing_value_generation_from_column(self):
        """Test transform when ``_missing_value_generation`` is set to ``from_column``.

        When ``missing_value_generation`` is 'from_column', the nulls should be replaced
        by the ``_missing_value_replacement`` and another column flagging the nulls
        should be created.
        """
        # Setup
        transformer = NullTransformer(missing_value_generation='from_column')
        transformer.nulls = False
        transformer._missing_value_replacement = 'c'
        input_data = pd.Series(['a', 'b', np.nan])

        # Run
        output = transformer.transform(input_data)

        # Assert
        expected_output = np.array([
            ['a', 0.0],
            ['b', 0.0],
            ['c', 1.0],
        ], dtype=object)
        np.testing.assert_equal(expected_output, output)

    def test_transform__missing_value_generation_random(self):
        """Test transform when ``_missing_value_generation`` is set to ``random``.

        Test that when the ``_missing_value_generation`` is set to ``random``, the nulls should
        be replaced by the ``_missing_value_replacement``.
        """
        # Setup
        transformer = NullTransformer()
        transformer._missing_value_replacement = 3
        input_data = pd.Series([1, 2, np.nan])

        # Run
        output = transformer.transform(input_data)

        # Assert
        expected_output = np.array([1, 2, 3])
        np.testing.assert_equal(expected_output, output)

        modified_input_data = pd.Series([1, 2, np.nan])
        pd.testing.assert_series_equal(modified_input_data, input_data)

    def test_reverse_transform__missing_value_generation_from_column_with_nulls(self):
        """Test reverse_transform when ``missing_value_generation`` is ``from_column`` and nulls.

        When ``missing_value_generation`` is ``from_column`` and there are nulls, the second column
        in the input data should be used to decide which values to replace with nan,
        by selecting the rows where the null column value is > 0.5.
        """
        # Setup
        transformer = NullTransformer(missing_value_generation='from_column')
        transformer.nulls = True
        input_data = np.array([
            [0.0, 0.0],
            [0.2, 0.2],
            [0.4, 0.4],
            [0.6, 0.6],
            [0.8, 0.8],
        ])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, np.nan, np.nan])
        pd.testing.assert_series_equal(expected_output, output)

    def test_reverse_transform__missing_value_generation_from_column_no_nulls(self):
        """Test reverse_transform when ``missing_value_generation`` is ``from_column``, no nulls.

        When ``missing_value_generation`` is ``from_column`` but no nulls are found, the second
        column of the input data must be dropped and the first one returned as a ``pd.Series``
        without having been modified.
        """
        # Setup
        transformer = NullTransformer(missing_value_generation='from_column')
        transformer.nulls = False
        input_data = np.array([
            [0.0, 0.0],
            [0.2, 0.2],
            [0.4, 0.4],
            [0.6, 0.6],
            [0.8, 0.8],
        ])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, 0.6, 0.8])
        pd.testing.assert_series_equal(expected_output, output)

    @patch('rdt.transformers.null.np.random')
    def test_reverse_transform__missing_value_generation_random_with_nulls(self, random_mock):
        """Test reverse_transform when ``missing_value_generation`` is ``random`` and nulls.

        When ``missing_value_generation`` is ``random`` and there are nulls, a ``_null_percentage``
        of values should randomly be replaced with ``np.nan``.
        """
        # Setup
        transformer = NullTransformer(missing_value_generation='random')
        transformer.nulls = True
        transformer._null_percentage = 0.5
        input_data = np.array([0.0, 0.2, 0.4, 0.6])
        random_mock.random.return_value = np.array([1, 1, 0, 1])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, np.nan, 0.6])
        pd.testing.assert_series_equal(expected_output, output)

    def test_reverse_transform__missing_value_generation_random_no_nulls(self):
        """Test reverse_transform when missing_value_generation is ``random`` and no nulls.

        When ``missing_value_generation`` is ``random`` and there are no nulls, the input data
        is not modified at all.
        """
        # Setup
        transformer = NullTransformer(missing_value_generation='random')
        transformer.nulls = False
        transformer._missing_value_replacement = 0.4
        input_data = np.array([0.0, 0.2, 0.4, 0.6])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, 0.6])
        pd.testing.assert_series_equal(expected_output, output)



import pytest

from rdt.transformers import (
    AnonymizedFaker, BinaryEncoder, FloatFormatter, LabelEncoder, RegexGenerator,
    UnixTimestampEncoder, get_default_transformers, get_transformer_class, get_transformer_name)
from rdt.transformers.addons.identity.identity import IdentityTransformer


def test_get_transformer_name():
    """Test the ``get_transformer_name`` method.

    Validate the method returns the class path when passed the class.

    Input:
        - a class.

    Output:
        - the path of the class.
    """
    # Setup
    transformer = BinaryEncoder

    # Run
    returned = get_transformer_name(transformer)

    # Assert
    assert returned == 'rdt.transformers.boolean.BinaryEncoder'


def test_get_transformer_name_incorrect_input():
    """Test the ``get_transformer_name`` method crashes.

    Validate the method raises a ``ValueError`` when passed a string.

    Input:
        - a string.

    Raises:
        - ``ValueError``, with the correct output message.
    """
    # Setup
    transformer = 'rdt.transformers.boolean.BinaryEncoder'

    # Run / Assert
    error_msg = 'The transformer rdt.transformers.boolean.BinaryEncoder must be passed as a class.'
    with pytest.raises(ValueError, match=error_msg):
        get_transformer_name(transformer)


def test_get_transformer_class_transformer_path():
    """Test the ``get_transformer_class`` method.

    Validate the method returns the correct class when passed the class path.

    Input:
        - a string describing the transformer path.

    Output:
        - the class corresponding to the transformer path.
    """
    # Setup
    transformer_path = 'rdt.transformers.boolean.BinaryEncoder'

    # Run
    returned = get_transformer_class(transformer_path)

    # Assert
    assert returned == BinaryEncoder


def test_get_transformer_class_partial_path():
    """Test with non fully specified path."""
    # Run
    returned = get_transformer_class('rdt.transformers.BinaryEncoder')

    # Assert
    assert returned == BinaryEncoder


def test_get_transformer_class_transformer_path_addon():
    """Test the ``get_transformer_class`` method.

    Validate the method returns the correct class when passed an addon path.

    Input:
        - a string describing the transformer path.

    Output:
        - the class corresponding to the transformer path.
    """
    # Setup
    transformer_path = 'rdt.transformers.addons.identity.identity.IdentityTransformer'

    # Run
    returned = get_transformer_class(transformer_path)

    # Assert
    assert returned == IdentityTransformer


def test_get_default_transformers():
    """Test the ``get_default_transformers`` method.

    Check that the right default transformer is returned for each type.
    """
    # Run
    default_transformer_dict = get_default_transformers()

    # Assert
    expected_dict = {
        'numerical': FloatFormatter,
        'categorical': LabelEncoder,
        'boolean': LabelEncoder,
        'datetime': UnixTimestampEncoder,
        'text': RegexGenerator,
        'pii': AnonymizedFaker,
    }

    for sdtype, transformer in expected_dict.items():
        assert isinstance(default_transformer_dict[sdtype], transformer)
